package com.tzld.piaoquan.recommend.server.implement;


import com.alibaba.fastjson.JSONObject;
import com.google.common.reflect.TypeToken;
import com.tzld.piaoquan.recommend.server.common.base.RankItem;
import com.tzld.piaoquan.recommend.server.framework.candidiate.Candidate;
import com.tzld.piaoquan.recommend.server.framework.common.User;
import com.tzld.piaoquan.recommend.server.framework.merger.MergeUtils;
import com.tzld.piaoquan.recommend.server.framework.merger.StrategyQueue;
import com.tzld.piaoquan.recommend.server.framework.recaller.BaseRecaller;
import com.tzld.piaoquan.recommend.server.framework.recaller.provider.RedisBackedQueue;
import com.tzld.piaoquan.recommend.server.framework.score.ScorerUtils;
import com.tzld.piaoquan.recommend.server.framework.utils.RedisSmartClient;
import com.tzld.piaoquan.recommend.server.gen.recommend.RecommendRequest;
import com.tzld.piaoquan.recommend.server.model.Video;
import com.tzld.piaoquan.recommend.server.service.rank.RankParam;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorItemFeature;
import com.tzld.piaoquan.recommend.server.service.rank.extractor.RankExtractorUserFeature;
import com.tzld.piaoquan.recommend.server.util.CommonCollectionUtils;
import com.tzld.piaoquan.recommend.server.util.JSONUtils;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Qualifier;
import org.springframework.data.redis.core.RedisTemplate;
import org.springframework.stereotype.Service;

import javax.annotation.Resource;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.stream.Collectors;

@Service
public class TopRecommendPipeline {

    private static final Logger log = LoggerFactory.getLogger(TopRecommendPipeline.class);

    public static final String FILTER_CONF = "filter_config.conf";
    public static final String MERGE_CONF = "merge_config.conf";

    public static final String PREFIX = "";

    @Resource
    private RedisSmartClient client;
    @Resource
    public RedisTemplate<String, String> redisTemplate;

    public List<RankItem> feedByRec(final RecommendRequest requestData,
                                    final int requestIndex,
                                    final User userInfo) {
        int recallNum =  200;

        // Step 1: Attention extraction
//        long timestamp = System.currentTimeMillis();
//        UserAttentionExtractorPipeline attentionExtractorPipeline = UserAttentionExtractorUtils.getAtttentionPipeline(UserAttentionExtractorUtils.BASE_CONF);
//        attentionExtractorPipeline.extractAttention(requestData, userInfo);

        // Step 2: create top queue
        StrategyQueue topQueue = MergeUtils.createTopQueue(MERGE_CONF, "top-queue");


        // Step 3: Candidate
        Map<String, Candidate> candidates = new HashMap<String, Candidate>();
        topQueue.candidate(candidates, recallNum, userInfo, requestData, 0, 0);


        // Step 4: Recalling & Basic Scoring
        RedisBackedQueue queueProvider = new RedisBackedQueue(client, 1000L);

        BaseRecaller recaller = new BaseRecaller(queueProvider);
        List<RankItem> items = recaller.recalling(requestData, userInfo, requestIndex, new ArrayList<>(candidates.values()));


        // Step 4: Advance Scoring
//        timestamp = System.currentTimeMillis();
//        ScorerPipeline scorerPipeline = getScorerPipeline(requestData);
//        items = scorerPipeline.scoring(requestData, userInfo, requestIndex, items);


        // Step 5: Merger
        MergeUtils.distributeItemsToMultiQueues(topQueue, items);
        topQueue.merge(recallNum * 3, userInfo, requestData, requestIndex, 0);

        // 多样性融合
//        List<RankItem> mergeItems = topQueue.getItems();
//        MergeUtils.diversityRerank(mergeItems, SimilarityUtils.getIsSameUserTagOrCategoryFunc(), recallNum, 6, 2);

        // Step 6: Global Rank & subList
        // TODO 前置和后置处理逻辑 hardcode，后续优化
        Map<String, String> sceneFeatureMap = getSceneFeature(requestData);
        Map<String, String> userFeatureMap = getUserFeatureMap(requestData, items);
        List<RankItem> rovRecallRankNewallScore = ScorerUtils.getScorerPipeline(ScorerUtils.BASE_CONF)
                .scoring(sceneFeatureMap, userFeatureMap, items);

        return rovRecallRankNewallScore;
    }

    private Map<String, String> getUserFeatureMap(RecommendRequest param, List<RankItem> rankItems) {
        Map<String, String> userFeatureMap = new HashMap<>(64);
        if (param.getMid() != null && !param.getMid().isEmpty()){
            String midKey = "user_info_4video_" + param.getMid();
            String userFeatureStr = redisTemplate.opsForValue().get(midKey);
            if (userFeatureStr != null){
                try{
                    userFeatureMap = JSONUtils.fromJson(userFeatureStr,
                            new TypeToken<Map<String, String>>() {},
                            userFeatureMap);
                }catch (Exception e){
                    log.error(String.format("parse user json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }else{
                JSONObject obj = new JSONObject();
                obj.put("name", "user_key_in_model_is_null");
                obj.put("class", this.getClass().getSimpleName());
                log.info(obj.toString());
//                return videos;
            }
        }
        final Set<String> userFeatureSet = new HashSet<>(Arrays.asList(
                "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system",
                "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
        ));
        Iterator<Map.Entry<String, String>> iterator = userFeatureMap.entrySet().iterator();
        while (iterator.hasNext()) {
            Map.Entry<String, String> entry = iterator.next();
            if (!userFeatureSet.contains(entry.getKey())) {
                iterator.remove();
            }
        }
        Map<String, String> f1 = RankExtractorUserFeature.getOriginFeature(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "machineinfo_brand", "machineinfo_model", "machineinfo_platform", "machineinfo_system"
                ))
        );
        Map<String, String> f2 = RankExtractorUserFeature.getUserRateFeature(userFeatureMap);
        Map<String, String> f3 = RankExtractorUserFeature.cntFeatureChange(userFeatureMap,
                new HashSet<String>(Arrays.asList(
                        "u_1day_exp_cnt", "u_1day_click_cnt", "u_1day_share_cnt", "u_1day_return_cnt",
                        "u_3day_exp_cnt", "u_3day_click_cnt", "u_3day_share_cnt", "u_3day_return_cnt"
                ))
        );
        f1.putAll(f2);
        f1.putAll(f3);
        log.info("userFeature in model = {}", JSONUtils.toJson(f1));

        // 2-1: item特征处理
        final Set<String> itemFeatureSet = new HashSet<>(Arrays.asList(
                "total_time", "play_count_total",
                "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"
        ));

        List<Long> videoIds = CommonCollectionUtils.toListDistinct(rankItems, RankItem::getVideoId);
        List<String> videoFeatureKeys = videoIds.stream().map(r-> "video_info_" + r)
                .collect(Collectors.toList());
        List<String> videoFeatures = redisTemplate.opsForValue().multiGet(videoFeatureKeys);
        if (videoFeatures != null){
            for (int i=0; i<videoFeatures.size(); ++i){
                String vF = videoFeatures.get(i);
                Map<String, String> vfMap = new HashMap<>();
                if (vF == null){
                    continue;
                }
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);
                    Map<String, String> vfMapCopy = new HashMap<>(vfMap);
                    rankItems.get(i).setItemBasicFeature(vfMapCopy);
                    Iterator<Map.Entry<String, String>> iteratorIn = vfMap.entrySet().iterator();
                    while (iteratorIn.hasNext()) {
                        Map.Entry<String, String> entry = iteratorIn.next();
                        if (!itemFeatureSet.contains(entry.getKey())) {
                            iteratorIn.remove();
                        }
                    }
                    Map<String, String> f4 = RankExtractorItemFeature.getItemRateFeature(vfMap);
                    Map<String, String> f5 = RankExtractorItemFeature.cntFeatureChange(vfMap,
                            new HashSet<String>(Arrays.asList(
                                    "total_time", "play_count_total",
                                    "i_1day_exp_cnt", "i_1day_click_cnt", "i_1day_share_cnt", "i_1day_return_cnt",
                                    "i_3day_exp_cnt", "i_3day_click_cnt", "i_3day_share_cnt", "i_3day_return_cnt"))
                    );
                    f4.putAll(f5);
                    rankItems.get(i).setFeatureMap(f4);
                }catch (Exception e){
                    log.error(String.format("parse video json is wrong in {} with {}", this.getClass().getSimpleName(), e));
                }
            }
        }
        // 2-2: item 实时特征处理
        List<String> rtFeaPartKey = new ArrayList<>(Arrays.asList("item_rt_fea_1day_partition", "item_rt_fea_1h_partition"));
        List<String> rtFeaPartKeyResult = this.redisTemplate.opsForValue().multiGet(rtFeaPartKey);
        Calendar calendar = Calendar.getInstance();
        String date = new SimpleDateFormat("yyyyMMdd").format(calendar.getTime());
        String hour = new SimpleDateFormat("HH").format(calendar.getTime());
        String rtFeaPart1day = date + hour;
        String rtFeaPart1h = date + hour;
        if (rtFeaPartKeyResult != null){
            if (rtFeaPartKeyResult.get(0) != null){
                rtFeaPart1day = rtFeaPartKeyResult.get(0);
            }
            if (rtFeaPartKeyResult.get(1) != null){
                rtFeaPart1h = rtFeaPartKeyResult.get(1);
            }
        }

        List<String> videoRtKeys1 = videoIds.stream().map(r-> "item_rt_fea_1day_" + r)
                .collect(Collectors.toList());
        List<String> videoRtKeys2 = videoIds.stream().map(r-> "item_rt_fea_1h_" + r)
                .collect(Collectors.toList());
        videoRtKeys1.addAll(videoRtKeys2);
        List<String> videoRtFeatures = this.redisTemplate.opsForValue().multiGet(videoRtKeys1);


        if (videoRtFeatures != null){
            int j = 0;
            for (RankItem item: rankItems){
                String vF = videoRtFeatures.get(j);
                ++j;
                if (vF == null){
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);
                    for (Map.Entry<String, String> entry : vfMap.entrySet()){
                        String value = entry.getValue();
                        if (value == null){
                            continue;
                        }
                        String [] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1){
                            String [] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1day_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1day);
                item.getFeatureMap().putAll(f8);
            }
            for (RankItem item: rankItems){
                String vF = videoRtFeatures.get(j);
                ++j;
                if (vF == null){
                    continue;
                }
                Map<String, String> vfMap = new HashMap<>();
                Map<String, Map<String, Double>> vfMapNew = new HashMap<>();
                try{
                    vfMap = JSONUtils.fromJson(vF, new TypeToken<Map<String, String>>() {}, vfMap);

                    for (Map.Entry<String, String> entry : vfMap.entrySet()){
                        String value = entry.getValue();
                        if (value == null){
                            continue;
                        }
                        String [] var1 = value.split(",");
                        Map<String, Double> tmp = new HashMap<>();
                        for (String var2 : var1){
                            String [] var3 = var2.split(":");
                            tmp.put(var3[0], Double.valueOf(var3[1]));
                        }
                        vfMapNew.put(entry.getKey(), tmp);
                    }
                    item.setItemRealTimeFeature(vfMapNew);
                }catch (Exception e){
                    log.error(String.format("parse video item_rt_fea_1h_ json is wrong in {} with {}",
                            this.getClass().getSimpleName(), e));
                }
                Map<String, String> f8 = RankExtractorItemFeature.getItemRealtimeRate(vfMapNew, rtFeaPart1h);
                item.getFeatureMap().putAll(f8);
            }
        }


        log.info("ItemFeature = {}", JSONUtils.toJson(videoFeatures));
        return userFeatureMap;
    }

    private Map<String, String> getSceneFeature(RecommendRequest param) {
        Map<String, String> sceneFeatureMap = new HashMap<>();
        String provinceCn = param.getProvince();
        provinceCn = provinceCn.replaceAll("省$", "");
        sceneFeatureMap.put("ctx_region", provinceCn);
        String city = param.getCity();
        if ("台北市".equals(city) |
                "高雄市".equals(city) |
                "台中市".equals(city) |
                "桃园市".equals(city) |
                "新北市".equals(city) |
                "台南市".equals(city) |
                "基隆市".equals(city) |
                "吉林市".equals(city) |
                "新竹市".equals(city) |
                "嘉义市".equals(city)
        ){
            ;
        }else{
            city = city.replaceAll("市$", "");
        }
        sceneFeatureMap.put("ctx_city", city);

        Calendar calendar = Calendar.getInstance();
        sceneFeatureMap.put("ctx_week", (calendar.get(Calendar.DAY_OF_WEEK) + 6) % 7 + "");
        sceneFeatureMap.put("ctx_hour", new SimpleDateFormat("HH").format(calendar.getTime()));

        return sceneFeatureMap;
    }




}
